from __future__ import division
from __future__ import absolute_import
import numpy as np

from rllab2.core.serializable import Serializable
from rllab2.envs.base import Step
from rllab2.envs.mujoco.mujoco_env import MujocoEnv
from rllab2.misc import autoargs
from rllab2.misc import logger
from rllab2.misc.overrides import overrides


def smooth_abs(x, param):
    return np.sqrt(np.square(x) + np.square(param)) - param


class Walker2DEnv(MujocoEnv, Serializable):

    FILE = u'walker2d.xml'

    @autoargs.arg(u'ctrl_cost_coeff', type=float,
                  help=u'cost coefficient for controls')
    def __init__(
            self,
            ctrl_cost_coeff=1e-2,
            *args, **kwargs):
        self.ctrl_cost_coeff = ctrl_cost_coeff
        super(Walker2DEnv, self).__init__(*args, **kwargs)
        Serializable.quick_init(self, locals())

    def get_current_obs(self):
        return np.concatenate([
            self.model.data.qpos.flat,
            self.model.data.qvel.flat,
            self.get_body_com(u"torso").flat,
        ])

    def step(self, action):
        self.forward_dynamics(action)
        next_obs = self.get_current_obs()
        action = np.clip(action, *self.action_bounds)
        lb, ub = self.action_bounds
        scaling = (ub - lb) * 0.5
        ctrl_cost = 0.5 * self.ctrl_cost_coeff * \
            np.sum(np.square(action / scaling))
        forward_reward = self.get_body_comvel(u"torso")[0]
        reward = forward_reward - ctrl_cost
        qpos = self.model.data.qpos
        done = not (qpos[0] > 0.8 and qpos[0] < 2.0
                    and qpos[2] > -1.0 and qpos[2] < 1.0)
        return Step(next_obs, reward, done)

    @overrides
    def log_diagnostics(self, paths):
        progs = [
            path[u"observations"][-1][-3] - path[u"observations"][0][-3]
            for path in paths
        ]
        logger.record_tabular(u'AverageForwardProgress', np.mean(progs))
        logger.record_tabular(u'MaxForwardProgress', np.max(progs))
        logger.record_tabular(u'MinForwardProgress', np.min(progs))
        logger.record_tabular(u'StdForwardProgress', np.std(progs))

